# Ozwillo plugin to deploy Publik
# Copyright (C) 2017  Entr'ouvert
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import logging
import requests
import json
import subprocess
import hmac
import threading
import tempfile
from hashlib import sha1

from django.views.decorators.csrf import csrf_exempt
from django.conf import settings
from django.http import (HttpResponseForbidden, HttpResponseBadRequest,
                         HttpResponseNotFound, HttpResponse)
from django.core.management import call_command
from django.utils.text import slugify

from .models import OzwilloInstance

logger = logging.getLogger(__name__)


def run_command(args):
    try:
        process = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    except OSError as e:
        logger.error('ozwillo: launching subprocess %s raised error %s', args, e)
        return False
    logger.debug('ozwillo: launching subprocess with pid %s : %s', process.pid, args)
    stdoutdata, stderrdata = process.communicate()
    if process.returncode != 0:
        logger.error('ozwillo: subprocess %s failed returncode=%s stdout=%r stderr=%r',
                     process.pid, process.returncode, stdoutdata, stderrdata)
        return False
    logger.debug('ozwillo: subprocess terminated')
    return True


def valid_signature_required(setting):
    '''Validate Ozwillo signatures'''
    signature_header_name = 'HTTP_X_HUB_SIGNATURE'

    def decorator(func):
        def wrapper(request, *args, **kwargs):
            api_secret = getattr(settings, setting)
            if signature_header_name in request.META:
                if request.META[signature_header_name].startswith('sha1='):
                    algo, received_hmac = request.META[signature_header_name].rsplit('=')
                    computed_hmac = hmac.new(api_secret, request.body, sha1).hexdigest()
                    # the received hmac is uppercase according to
                    # http://doc.ozwillo.com/#ref-3-2-1
                    if received_hmac.lower() != computed_hmac:
                        logger.error(u'ozwillo: invalid HMAC')
                        return HttpResponseForbidden('invalid HMAC')
                else:
                    logger.error(u'ozwillo: invalid HMAC algo')
                    return HttpResponseForbidden('invalid HMAC algo')
            else:
                logger.error(u'ozwillo: no HMAC in the header')
                return HttpResponseForbidden('no HMAC in the header')
            return func(request, *args, **kwargs)
        return wrapper
    return decorator


def is_ozwillo_enabled(func):
    def wrapper(request):
        if not getattr(settings, 'OZWILLO_ENABLED', False):
            return HttpResponseNotFound('owillo providing is not active here.')
        return func(request)
    return wrapper


@csrf_exempt
@is_ozwillo_enabled
@valid_signature_required(setting='OZWILLO_SECRET')
def create_publik_instance(request):
    try:
        data = json.loads(request.body)
    except ValueError:
        logger.warning(u'ozwillo: received non JSON request')
        return HttpResponseBadRequest('invalid JSON content')

    logger.debug(u'ozwillo: create publik instance request, %r', data)

    if 'organization_name' not in data.keys():
        logger.warning(u'ozwillo: missing organization_name')
        return HttpResponseBadRequest('missing parameter "organization_name"')

    org_name = slugify(data['organization_name'])
    if OzwilloInstance.objects.filter(domain_slug=org_name):
        logger.warning(u'ozwillo: instance %s already exists', org_name)
        return HttpResponseBadRequest('instance %s already exists' % org_name)

    OzwilloInstance.objects.create(external_ozwillo_id=data['instance_id'], domain_slug=org_name)

    def thread_function(data):
        try:
            ozwillo_deploy_thread(data)
        except Exception:
            logger.exception(u'ozwillo: error occured while deploying instance %s',
                             data['organization_name'])

    thread = threading.Thread(target=thread_function, args=(data,))
    thread.start()

    return HttpResponse()


def ozwillo_deploy_thread(data):
    logger.debug(u'ozwillo: deploy thread start')
    # Request parsing
    client_id = data['client_id']
    client_secret = data['client_secret']
    instance_id = data['instance_id']
    instance_name = data['organization_name']
    instance_name = slugify(instance_name)
    registration_uri = data['instance_registration_uri']
    user = data['user']

    # Cook new platform
    template_recipe = json.load(open('/etc/hobo/ozwillo/template_recipe.json', 'rb'))
    var = template_recipe['variables']
    for key, value in var.items():
        var[key] = value.replace('instance_name', instance_name)

    template_recipe['variables'] = var
    domain = var['combo']
    domain_agent = var['combo_agent']
    domain_passerelle = var['passerelle']
    logger.debug(u'ozwillo: cooking %s', template_recipe)

    with tempfile.NamedTemporaryFile() as recipe_file:
        json.dump(template_recipe, recipe_file)
        recipe_file.flush()
        call_command('cook', recipe_file.name, timeout=1000, verbosity=0)

    # Load user portal template
    logger.debug(u'ozwillo: loading combo template')
    run_command([
        'sudo', '-u', 'combo',
        'combo-manage', 'tenant_command', 'import_site',
        '/etc/hobo/ozwillo/import-site-template.json',
        '-d', domain
    ])

    # Load agent portal template
    logger.debug(u'ozwillo: loading combo agent template')
    run_command([
        'sudo', '-u', 'combo',
        'combo-manage', 'tenant_command', 'import_site',
        '/etc/hobo/ozwillo/import-site-agents.json',
        '-d', domain_agent
    ])

    # Configure OIDC Ozwillo authentication
    logger.debug(u'ozwillo: configuring OIDC ozwillo authentication')
    domain_name = 'connexion-%s.%s' % (instance_name, settings.OZWILLO_ENV_DOMAIN)
    if run_command([
            'sudo', '-u', 'authentic-multitenant',
            'authentic2-multitenant-manage', 'tenant_command', 'oidc-register-issuer',
            '-d', domain_name,
            '--scope', 'profile',
            '--scope', 'email',
            '--issuer', settings.OZWILLO_PLATEFORM,
            '--client-id', client_id,
            '--client-secret', client_secret,
            '--claim-mapping', 'given_name first_name always_verified',
            '--claim-mapping', 'family_name last_name always_verified',
            '--ou-slug', 'default',
            '--claim-mapping', 'email email required',
            'Ozwillo'
    ]):
        # creation of the admin user depends upon the creation of provider
        logger.debug(u'ozwillo: creating admin user')
        create_user_script = os.path.dirname(__file__) + '/scripts/create_user_ozwillo.py'
        run_command([
            'sudo', '-u', 'authentic-multitenant',
            'authentic2-multitenant-manage', 'tenant_command', 'runscript', '-d', domain_name,
            create_user_script, user['email_address'], user['id'], user['name']
        ])

    # Load passerelle template
    logger.debug(u'ozwillo: loading passerelle template')
    run_command([
        'sudo', '-u', 'passerelle',
        'passerelle-manage', 'tenant_command', 'import_site',
        '/etc/hobo/ozwillo/import-site-passerelle.json',
        '--import-user', '-d', domain_passerelle
    ])

    # Sending done event to Ozwillo
    services = {
        'services': [{
            'local_id': 'publik',
            'name': 'Publik - %s' % (instance_name),
            'service_uri': 'https://connexion-%s.%s/accounts/oidc/login?iss=%s'
                           % (instance_name, settings.OZWILLO_ENV_DOMAIN,
                              settings.OZWILLO_PLATEFORM),
            'description': 'Gestion de la relation usagers',
            'tos_uri': 'https://publik.entrouvert.com/',
            'policy_uri': 'https://publik.entrouvert.com/',
            'icon': 'https://publik.entrouvert.com/static/img/logo-publik-64x64.png',
            'payment_option': 'FREE',
            'target_audience': ['PUBLIC_BODIES',
                                'CITIZENS',
                                'COMPANIES'],
            'contacts': ['https://publik.entrouvert.com/'],
            'redirect_uris': ['https://connexion-%s.%s/accounts/oidc/callback/'
                              % (instance_name, settings.OZWILLO_ENV_DOMAIN)],
        }],
        'instance_id': instance_id,
        'destruction_uri': settings.OZWILLO_DESTRUCTION_URI,
        'destruction_secret': settings.OZWILLO_DESTRUCTION_SECRET,
        'needed_scopes': []
    }
    logger.debug(u'ozwillo: sending registration request, %r', services)
    headers = {'Content-type': 'application/json', 'Accept': 'application/json'}
    response = requests.post(
        registration_uri,
        data=json.dumps(services),
        auth=(client_id, client_secret),
        headers=headers)
    logger.debug(u'ozwillo: registration response, status=%s content=%r', response.status_code,
                 response.content)
    logger.debug(u'ozwillo: deploy thread finished')


@csrf_exempt
@is_ozwillo_enabled
@valid_signature_required(setting='OZWILLO_DESTRUCTION_SECRET')
def delete_publik_instance(request):
    try:
        data = json.loads(request.body)
    except ValueError:
        logger.warning(u'ozwillo: received non JSON request')
        return HttpResponseBadRequest('invalid JSON content')

    logger.debug(u'ozwillo: delete publik instance request, %r', data)

    try:
        instance = OzwilloInstance.objects.get(external_ozwillo_id=data['instance_id'])
    except OzwilloInstance.DoesNotExist:
        return HttpResponseBadRequest('no instance with id %s' % data['instance_id'])

    def thread_function(data):
        try:
            ozwillo_destroy_thread(instance)
            logger.debug(u'ozwillo: instance %s destroyed', instance.domain_slug)
        except Exception:
            logger.exception('ozwillo: error occured while destroying instance %s',
                             instance.domain_slug)
    thread = threading.Thread(target=ozwillo_destroy_thread, args=(instance,))
    thread.start()

    return HttpResponse(status=200)


def ozwillo_destroy_thread(instance):
    logger.debug(u'ozwillo: destroy thread start')
    instance_slug = instance.domain_slug
    services = settings.OZWILLO_SERVICES
    wcs = services['wcs-au-quotidien']

    for s, infos in services.items():
        # to get the two combo instances which have same service
        service_user = s.split('_')[0]

        tenant = '%s%s.%s' % (infos[0], instance_slug, settings.OZWILLO_ENV_DOMAIN)

        run_command([
            'sudo', '-u', service_user, infos[1],
            'delete_tenant', '--force-drop', tenant
        ])

    tenant = '%s%s.%s' % (wcs[0], instance_slug, settings.OZWILLO_ENV_DOMAIN)
    run_command([
        'sudo', '-u', 'wcs',
        wcs[1], '-f', '/etc/wcs/wcs-au-quotidien.cfg',
        'delete_tenant', '--force-drop', tenant
    ])

    instance.delete()
    logger.debug(u'ozwillo: destroy thread finished')
